/*
   +----------------------------------------------------------------------+
   | PHP Version 5                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2007 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author:  Ard Biesheuvel <a.k.biesheuvel@its.tudelft.nl>              |
   +----------------------------------------------------------------------+
 */

/* $Id: interbase.rc 242949 2007-09-26 15:44:16Z cvs2svn $ */

#ifdef APSTUDIO_INVOKED
#error This file cannot be opened from the Visual Studio IDE
#endif

#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
#include "winres.h"
#undef APSTUDIO_READONLY_SYMBOLS

#include "php_version.h"

#ifdef _DEBUG
# define BUILD_TYPE "Debug"
#else
# define BUILD_TYPE "Release"
#endif

#include <ibase.h>
#include "interbase.c"

#ifdef FB_SQLDA
#define CLIENT_LIB "fbclient.dll"
#else
#define CLIENT_LIB "gds32.dll"
#endif

#ifndef _MAC

VS_VERSION_INFO VERSIONINFO
 FILEVERSION PHP_MAJOR_VERSION,PHP_MINOR_VERSION,PHP_RELEASE_VERSION,0
 PRODUCTVERSION PHP_MAJOR_VERSION,PHP_MINOR_VERSION,PHP_RELEASE_VERSION,0
#ifndef _DEBUG
 FILEFLAGS 0x0L
#else
 FILEFLAGS 0x1L
#endif
 FILEFLAGSMASK 0x3fL
 FILEOS VOS__WINDOWS32
 FILETYPE VFT_DLL
 FILESUBTYPE VFT2_UNKNOWN
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "000004b0"
        BEGIN
            VALUE "FileDescription", "PHP Interbase module\0"
            VALUE "FileVersion", PHP_VERSION "\0"
            VALUE "LegalCopyright", "Copyright © 2003 The PHP Group\0"
			VALUE "Build Type", BUILD_TYPE "\0"
            VALUE "Client Library", CLIENT_LIB "\0"
			VALUE "URL", "http://www.php.net\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x0, 1200
    END
END

#endif    // !_MAC
