EXPORTS
sqlite_open
sqlite_close
sqlite_exec
sqlite_last_insert_rowid
sqlite_changes
sqlite_error_string
sqlite_interrupt
sqlite_complete
sqlite_busy_handler
sqlite_busy_timeout
sqlite_get_table
sqlite_free_table
sqlite_exec_printf
sqlite_exec_vprintf
sqlite_get_table_printf
sqlite_get_table_vprintf
sqlite_mprintf
sqlite_vmprintf
sqlite_freemem
sqlite_libversion
sqlite_libencoding
sqlite_create_function
sqlite_create_aggregate
sqlite_function_type
sqlite_set_result_string
sqlite_set_result_int
sqlite_set_result_double
sqlite_set_result_error
sqlite_user_data
sqlite_aggregate_context
sqlite_aggregate_count
sqlite_set_authorizer
sqlite_trace
sqlite_compile
sqlite_step
sqlite_finalize
; some experimental stuff
sqlite_last_statement_changes
sqlite_reset
sqlite_bind
sqlite_progress_handler
sqlite_commit_hook
